@@ -14,6 +14,9 @@ import { differenceInMilliseconds } from 'date-fns';
 export const PLUGIN_NAME = 'jupyterlab-execute-time';
 const EXECUTE_TIME_CLASS = 'execute-time';
 
+const TOOLTIP_PREFIX = 'Previous Runs:';
+const PREV_DATA_EXECUTION_TIME_ATTR = 'data-prev-execution-time';
+
 // How long do we animate the color for
 const ANIMATE_TIME_MS = 1000;
 const ANIMATE_CSS = `executeHighlight ${ANIMATE_TIME_MS}ms`;
@@ -23,6 +26,9 @@ export interface IExecuteTimeSettings {
   highlight: boolean;
   positioning: string;
   minTime: number;
+  textContrast: string;
+  showLiveExecutionTime: boolean;
+  historyCount: number;
 }
 
 export default class ExecuteTimeWidget extends Widget {
@@ -34,12 +40,29 @@ export default class ExecuteTimeWidget extends Widget {
     super();
     this._panel = panel;
     this._tracker = tracker;
+    this._settingRegistry = settingRegistry;
 
     this.updateConnectedCell = this.updateConnectedCell.bind(this);
     settingRegistry.load(`${PLUGIN_NAME}:settings`).then(
       (settings: ISettingRegistry.ISettings) => {
         this._updateSettings(settings);
         settings.changed.connect(this._updateSettings.bind(this));
+
+        // If the plugin is enabled, force recoding of timing
+        // We only do this once (not on every settings update) in case the user tries to trun it off
+        if (settings.get('enabled').composite) {
+          this._settingRegistry
+            .load('@jupyterlab/notebook-extension:tracker')
+            .then(
+              (nbSettings: ISettingRegistry.ISettings) =>
+                nbSettings.set('recordTiming', true),
+              (err: Error) => {
+                console.error(
+                  `jupyterlab-execute-time: Could not force metadata recording: ${err}`
+                );
+              }
+            );
+        }
       },
       (err: Error) => {
         console.error(
@@ -142,6 +165,13 @@ export default class ExecuteTimeWidget extends Widget {
 
       if (!executionTimeNode) {
         executionTimeNode = document.createElement('div') as HTMLDivElement;
+        executionTimeNode.appendChild(document.createElement('span'));
+        // Use this over gap as hover is not a flexbox
+        const spacer = document.createElement('div');
+        spacer.style.minWidth = '12px';
+        executionTimeNode.appendChild(spacer);
+        executionTimeNode.appendChild(document.createElement('span'));
+
         if (!cell.inputHidden) {
           parentNode.append(executionTimeNode);
         }
@@ -175,7 +205,8 @@ export default class ExecuteTimeWidget extends Widget {
           );
       }
       const positioningClass = `${EXECUTE_TIME_CLASS}-positioning-${this._settings.positioning}`;
-      executionTimeNode.className = `${EXECUTE_TIME_CLASS} ${positioningClass}`;
+      const textContrastClass = `${EXECUTE_TIME_CLASS}-contrast-${this._settings.textContrast}`;
+      executionTimeNode.className = `${EXECUTE_TIME_CLASS} ${positioningClass} ${textContrastClass}`;
 
       // More info about timing: https://jupyter-client.readthedocs.io/en/stable/messaging.html#messages-on-the-shell-router-dealer-channel
       // A cell is queued when the kernel has received the message
@@ -211,18 +242,75 @@ export default class ExecuteTimeWidget extends Widget {
           this._settings.minTime <=
           differenceInMilliseconds(endTime, startTime) / 1000.0
         ) {
-          msg = `Last executed at ${getTimeString(endTime)} in ${getTimeDiff(
-            endTime,
-            startTime
-          )}`;
+          const executionTime = getTimeDiff(endTime, startTime);
+          const lastExecutionTime = executionTimeNode.getAttribute(
+            PREV_DATA_EXECUTION_TIME_ATTR
+          );
+          // Store the last execution time in the node to be used for various options
+          executionTimeNode.setAttribute(
+            PREV_DATA_EXECUTION_TIME_ATTR,
+            executionTime
+          );
+          // Only add a tooltip for all non-displayed execution times.
+          if (this._settings.historyCount > 0 && lastExecutionTime) {
+            let tooltip = executionTimeNode.getAttribute('title');
+            const executionTimes = [lastExecutionTime];
+            if (tooltip) {
+              executionTimes.push(
+                ...tooltip.substring(TOOLTIP_PREFIX.length + 1).split('\n')
+              );
+              // JS does the right thing of having empty items if extended
+              executionTimes.length = this._settings.historyCount;
+            }
+            tooltip = `${TOOLTIP_PREFIX}\n${executionTimes.join('\n')}`;
+            executionTimeNode.setAttribute('title', tooltip);
+          }
+          executionTimeNode.children[2].textContent = '';
+
+          msg = `Last executed at ${getTimeString(
+            endTime
+          )} in ${executionTime}`;
         }
       } else if (startTime) {
+        if (this._settings.showLiveExecutionTime) {
+          const lastRunTime = executionTimeNode.getAttribute(
+            'data-prev-execution-time'
+          );
+          const workingTimer = setInterval(() => {
+            if (
+              !executionTimeNode.children[0].textContent.startsWith(
+                'Execution started at'
+              )
+            ) {
+              clearInterval(workingTimer);
+              return;
+            }
+            if (
+              this._settings.minTime <=
+              differenceInMilliseconds(new Date(), startTime) / 1000.0
+            ) {
+              const executionTime = getTimeDiff(new Date(), startTime);
+
+              executionTimeNode.children[2].textContent = `${executionTime} ${
+                lastRunTime ? `(${lastRunTime})` : ''
+              }`;
+            }
+          }, 100);
+        }
         msg = `Execution started at ${getTimeString(startTime)}`;
       } else if (queuedTime) {
+        const lastRunTime = executionTimeNode.getAttribute(
+          'data-prev-execution-time'
+        );
+        if (this._settings.showLiveExecutionTime && lastRunTime) {
+          executionTimeNode.children[2].textContent = `N/A (${lastRunTime})`;
+        }
+
         msg = `Execution queued at ${getTimeString(queuedTime)}`;
       }
-      if (executionTimeNode.innerText !== msg) {
-        executionTimeNode.innerText = msg;
+      if (executionTimeNode.textContent !== msg) {
+        executionTimeNode.children[0].textContent = msg;
+
         if (!disableHighlight && this._settings.highlight && endTimeStr) {
           executionTimeNode.style.setProperty('animation', ANIMATE_CSS);
           setTimeout(
@@ -232,8 +320,14 @@ export default class ExecuteTimeWidget extends Widget {
         }
       }
     } else {
-      // Clean up in case it was removed
-      this._removeExecuteNode(cell);
+      // Hide it if data was removed (e.g. clear output).
+      // Don't remove as element store history, which are useful for later showing past runtime.
+      const executionTimeNode = cell.node.querySelector(
+        `.${EXECUTE_TIME_CLASS}`
+      );
+      if (executionTimeNode) {
+        executionTimeNode.classList.add('execute-time-hidden');
+      }
     }
   }
 
@@ -243,6 +337,12 @@ export default class ExecuteTimeWidget extends Widget {
     this._settings.positioning = settings.get('positioning')
       .composite as string;
     this._settings.minTime = settings.get('minTime').composite as number;
+    this._settings.textContrast = settings.get('textContrast')
+      .composite as string;
+    this._settings.showLiveExecutionTime = settings.get('showLiveExecutionTime')
+      .composite as boolean;
+    this._settings.historyCount = settings.get('historyCount')
+      .composite as number;
 
     const cells = this._panel.context.model.cells;
     if (this._settings.enabled) {
@@ -266,10 +366,14 @@ export default class ExecuteTimeWidget extends Widget {
   } = {};
   private _tracker: INotebookTracker;
   private _panel: NotebookPanel;
+  private _settingRegistry: ISettingRegistry;
   private _settings: IExecuteTimeSettings = {
     enabled: false,
     highlight: true,
     positioning: 'left',
     minTime: 0,
+    textContrast: 'high',
+    showLiveExecutionTime: true,
+    historyCount: 5,
   };
 }
